/*
*Copyright(c)2014, Jeffrey Lee
*Allrightsreserved.
*
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
*/
#ifndef COMMON_H
#define COMMON_H

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdint.h>
#include <ctype.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include <algorithm>
#include <vector>
#include <string>
#include <map>

//#define EXTRAVERBOSE /* Enable this for -v to give lots more output */

#define YED /* Output (minimal) GraphML suitable for yEd */

#define PROMOTE_CHILDREN /* Reduce tree size by promoting the tests children perform */

#define MAX_FUNC_PARAMS 2 /* Used in an optimisation or two */

using namespace std;

typedef float infogain;

typedef struct {
	uint8_t startbit; /* Start (lowest) bit */
	uint8_t bitcount; /* Size of field in bits */
	bool relevant; /* True if the field is relevant to the action */
	const char *name; /* Name used in constraints/code */
} field;

typedef enum {
	EFIELD,
	ECONST,
	EFUNC,
} cntype; /* Constraint type */

typedef enum
{
	ECODE, /* Value is block of code */
	ESTRING, /* Value is a quoted string */
	ESTRINGREF, /* Value is a variable name to use for a string arg */
} tokentype;

typedef struct
{
	tokentype type;
	string value;
} token;

typedef struct
{
	char *name; /* Minus the parameter list */
	char *value; /* With parameters converted to char codes 0x01-0x1f */
	size_t namelen;
	uint32_t numparam;
} macro;

class funcdef;
class constraint;
class actiondef;
class node;
class encoding;
class opcode_subset;

typedef vector<encoding *> encodingset;
typedef vector<actiondef *> actionset;
typedef vector<const constraint *> constraintset;
typedef vector<macro> macroset;
typedef vector<token> tokenset;

extern encodingset allencodings;
extern actionset allactions;

extern uint32_t opcodesize;
extern uint32_t opcodebits;

extern FILE *fout;
extern FILE *fgraph;
extern FILE *fcache;

extern encodingset allencodings;
extern actionset allactions;
extern bool verbose;
extern uint32_t maxmaskbits;
extern bool fuzzy;
extern bool incomplete;
extern bool valid;
extern const char *prefix;
extern bool isvoid;
extern uint32_t maxpreextract;
extern bool treewalker;
extern bool useargstruct;
extern bool readcache;
extern bool writecache;
extern bool updatecache;
extern bool badcache;
extern constraintset cacheconstraints;

extern bool cache_read(void *p,size_t s);
extern void cache_write(const void *p,size_t s);

#endif
