/*
*Copyright(c)2014, Jeffrey Lee
*Allrightsreserved.
*
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
*/
#ifndef ENCODING_H
#define ENCODING_H

#include "common.h"

class encoding
{
private:
	encoding();
public:
	const char *name; /* Name of encoding/action */
	const char *text; /* Definition from encodings file (fields only) */
	uint32_t fixed; /* Fixed bits of encoding */
	uint32_t mask; /* Mask of which bits are fixed (1=fixed, 0=variable) */
	uint32_t numfields; /* Number of variable fields */
	field *fields; /* Array of fields */
	uint32_t numconstraints; /* Number of constraints */
	constraint **constraints; /* Array of constraint ptrs */
	uint32_t shouldmatch_value; /* Value which the 'should match' bits should match */
	uint32_t shouldmatch_mask; /* Mask of which bits should match shouldmatch_value */

	const actiondef *action; /* Action to undertake on match */
	const encoding *kin; /* Encoding (if any) for which samecode() == true; null for root kin */
	const encoding *kin_next; /* Linked list of kin (head of list contained in root kin) */ 

	~encoding();

	bool matches(uint32_t opcode) const; /* True/false for whether matches opcode */
	uint32_t getfields() const; /* Get mask of constrained bits */

	void emit_comment() const;

	bool samecode(const encoding *e) const; /* Returns true if both encodings can utilise the same code (actions match, and relevant fields are identical) */
	bool samefields(const encoding *e) const; /* Similar to above, but only checks if both actions take the same parameters */

	string describe() const; /* Return string describing encoding, virtually identical to as in source file */

	void subtract(const encoding *e); /* Modify this encoding so that it doesn't match against any of e. May only work correctly if e lies fully within this? */

	static encoding *parse(const macroset &macros);

	static encoding *generateinverse(const encodingset &e,const char *name,const char *text); /* Generate an encoding that matches the inverse of an encodingset */
};

#endif
