/*
*Copyright(c)2014, Jeffrey Lee
*Allrightsreserved.
*
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
*/

#include "funcdef.h"

funcdef::funcdef(const char *n,uint32_t np,funcdeffunc f,const char *c,bool ni,bool si,bool no)
{
	assert(np<=MAX_FUNC_PARAMS);
	name = n;
	numparam = np;
	func = f;
	code = c;
	emitted = false;
	numeric_input = ni;
	same_input = si | ni; /* The checks we perform require same_input to be true if numeric_input is */
	numeric_output = no;
	treewalk_funcid = -1;
}

#define FUNCDEF(X,N,C,NI,SI,NO) funcdef(#X,N,func_ ## X,C,NI,SI,NO)

#define INFIXDEF(X,O) uint32_t func_ ## X(const uint32_t *p) { return (p[0] O p[1]); }

INFIXDEF(eq,==)
INFIXDEF(ne,!=)
INFIXDEF(lt,<)
INFIXDEF(gt,>)
INFIXDEF(le,<=)
INFIXDEF(ge,>=)
INFIXDEF(add,+)
INFIXDEF(sub,-)
INFIXDEF(mul,*)
INFIXDEF(div,/)
INFIXDEF(mod,%)
INFIXDEF(land,&&)
INFIXDEF(lor,||)
uint32_t func_lnot(const uint32_t *p) { return !p[0]; }
uint32_t func_leor(const uint32_t *p) { return (((p[0]!=0) ^ (p[1]!=0))!=0); }
INFIXDEF(band,&)
INFIXDEF(bor,|)
uint32_t func_bnot(const uint32_t *p) { return ~p[0]; }
INFIXDEF(beor,^)
INFIXDEF(lsl,<<)
INFIXDEF(lsr,>>)
uint32_t func_neg(const uint32_t *p) { return -p[0]; }
uint32_t func_bitcount(const uint32_t *p)
{
	/* Yes, this works. */
	uint32_t bits = p[0];
	bits -= (bits & 0xaaaaaaaa)>>1;
	bits = (bits & 0x33333333) + ((bits>>2) & 0x33333333);
	bits += bits>>4;
	bits &= 0x0f0f0f0f;
	bits += bits>>8;
	bits += bits>>16;
	return (bits & 0xff);
}
uint32_t func_bit(const uint32_t *p)
{
	return ((uint32_t)1)<<p[0];
}
uint32_t func_bits(const uint32_t *p)
{
	/* Syntax is high,low (all inclusive) */
	return (((uint32_t)1)<<(p[0]+1))-((((uint32_t)1)<<p[1])-1);
}
uint32_t func_ispow2(const uint32_t *p)
{
	return !(p[0] & (p[0]-1));
}
uint32_t func_BadReg(const uint32_t *p)
{
	return ((p[0] == 13) || (p[0] == 15));
}

funcdef funcdefs[] = {
	/* Note: these first four must be in the order below! */
	FUNCDEF(eq,2,"#define %sfunc_eq(X,Y) (X)==(Y)\n",false,true,false),
	FUNCDEF(land,2,"#define %sfunc_land(X,Y) (X)&&(Y)\n",false,false,false),
	FUNCDEF(lnot,1,"#define %sfunc_lnot(X) !(X)\n",false,false,false),
	FUNCDEF(lor,2,"#define %sfunc_lor(X,Y) (X)||(Y)\n",false,false,false),
	FUNCDEF(ne,2,"#define %sfunc_ne(X,Y) (X)!=(Y)\n",false,true,false),
	FUNCDEF(lt,2,"#define %sfunc_lt(X,Y) (X)<(Y)\n",true,true,false),
	FUNCDEF(gt,2,"#define %sfunc_gt(X,Y) (X)>(Y)\n",true,true,false),
	FUNCDEF(le,2,"#define %sfunc_le(X,Y) (X)<=(Y)\n",true,true,false),
	FUNCDEF(ge,2,"#define %sfunc_ge(X,Y) (X)>=(Y)\n",true,true,false),
	FUNCDEF(add,2,"#define %sfunc_add(X,Y) (X)+(Y)\n",true,true,true),
	FUNCDEF(sub,2,"#define %sfunc_sub(X,Y) (X)-(Y)\n",true,true,true),
	FUNCDEF(mul,2,"#define %sfunc_mul(X,Y) (X)*(Y)\n",true,true,true),
	FUNCDEF(div,2,"#define %sfunc_div(X,Y) (X)/(Y)\n",true,true,true),
	FUNCDEF(mod,2,"#define %sfunc_mod(X,Y) (X)%%(Y)\n",true,true,true),
	FUNCDEF(leor,2,"#define %sfunc_leor(X,Y) ((X)!=0) ^ ((Y)!=0)\n",false,false,false),
	FUNCDEF(band,2,"#define %sfunc_band(X,Y) (X)&(Y)\n",true,true,true),
	FUNCDEF(bor,2,"#define %sfunc_bor(X,Y) (X)|(Y)\n",true,true,true),
	FUNCDEF(bnot,1,"#define %sfunc_bnot(X) ~(X)\n",true,true,true),
	FUNCDEF(beor,2,"#define %sfunc_beor(X,Y) (X)^(Y)\n",true,true,true),
	FUNCDEF(lsl,2,"#define %sfunc_lsl(X,Y) (X)<<(Y)\n",true,true,true),
	FUNCDEF(lsr,2,"#define %sfunc_lsr(X,Y) (X)>>(Y)\n",true,true,true),
	FUNCDEF(neg,1,"#define %sfunc_neg(X) -(X)\n",true,true,true),
	FUNCDEF(bitcount,1,"\
static uint32_t %sfunc_bitcount(uint32_t bits)\n\
{\n\
	bits -= (bits & 0xaaaaaaaa)>>1;\n\
	bits = (bits & 0x33333333) + ((bits>>2) & 0x33333333);\n\
	bits += bits>>4;\n\
	bits &= 0x0f0f0f0f;\n\
	bits += bits>>8;\n\
	bits += bits>>16;\n\
	return (bits & 0x1f);\n\
}\n",true,true,true),
	FUNCDEF(bit,1,"#define %sfunc_bit(X) ((uint32_t)1)<<(X)\n",true,true,true),
	FUNCDEF(bits,2,"#define %sfunc_bits(X,Y) (((uint32_t)1)<<((X)+1))-((((uint32_t)1)<<(Y))-1)\n",true,true,true),
	FUNCDEF(ispow2,1,"#define %sfunc_ispow2(X) !((X) & ((X)-1))\n",true,true,false),
	FUNCDEF(BadReg,1,"#define %sfunc_BadReg(X) (((X)==13)||((X)==15))\n",true,true,false),
};

const uint32_t numfuncdefs = sizeof(funcdefs)/sizeof(funcdef);

/* This is why the first four must be in order */
funcdef *funcdef_eq = &funcdefs[0];
funcdef *funcdef_land = &funcdefs[1];
funcdef *funcdef_lnot = &funcdefs[2];
funcdef *funcdef_lor = &funcdefs[3];

void funcdef::emit()
{
	if(emitted)
		return;
	emitted = true;
	fprintf(fout,"#ifndef %sfunc_%s\n",prefix,name);
	fprintf(fout,code,prefix);
	fprintf(fout,"#endif\n");
}

funcdef *funcdef_get(const char *name)
{
	for(size_t i=0;i<numfuncdefs;i++)
		if(!strcmp(funcdefs[i].getname(),name))
			return &funcdefs[i];
	return NULL;
}
