/*
*Copyright(c)2014, Jeffrey Lee
*Allrightsreserved.
*
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
*/

#include "common.h"

tokenset tokenise(const char *code)
{
	/* "tokenise" a block of C code
	   At the moment we only deal with two token types (code and string
	   literals), so this is pretty straightforward */
	tokenset ret;
	while(*code)
	{
		token t;
		if(*code == '"')
		{
			t.type = ESTRING;
			t.value.push_back(*code++);
			do
			{
				if(*code == '\\')
				{
					t.value.push_back(*code++);
					if(*code)
						t.value.push_back(*code++);
				}
				else if(*code == '"')
				{
					t.value.push_back(*code++);
					ret.push_back(t);
					break;
				}
				else if(*code)
				{
					t.value.push_back(*code++);
				}
				else
				{
					/* Bad string! Push it anyway */
					ret.push_back(t);
					break;
				}
			} while(1);
		}
		else
		{
			t.type = ECODE;
			while(*code && (*code != '"'))
			{
				if((code[0] == '/') && (code[1] == '/'))
				{
					/* Single-line comment */
					t.value.push_back(*code++);
					t.value.push_back(*code++);
					while(*code && (*code != '\r') && (*code != '\n'))
					{
						if(*code == '\\')
						{
							t.value.push_back(*code++);
							/* Detect pesky CR-LF combinations */
							if(((code[0] == '\r') && (code[1] == '\n')) || ((code[0] == '\n') && (code[1] == '\r')))
							{
								t.value.push_back(*code++);
								t.value.push_back(*code++);
							}
							else if(*code)
								t.value.push_back(*code++);
						}
						else
							t.value.push_back(*code++);
					}
				}
				else if((code[0] == '/') && (code[1] == '*'))
				{
					/* Multi-line comment */
					t.value.push_back(*code++);
					t.value.push_back(*code++);
					while(*code && !((code[0] == '*') && (code[1] == '/')))
					{
						if(*code == '\\')
						{
							t.value.push_back(*code++);
							if(*code)
								t.value.push_back(*code++);
						}
						else
							t.value.push_back(*code++);
					}
				}
				else if(*code == '\\')
				{
					t.value.push_back(*code++);
					if(*code)
						t.value.push_back(*code++);
				}
				else
					t.value.push_back(*code++);
			}
			ret.push_back(t);
		}
	}
	return ret;
}

bool compatible_tokensets(const tokenset &a,const tokenset &b)
{
	if(a.size() != b.size())
		return false;
	for(size_t i=0;i<a.size();i++)
	{
		const token &ta = a[i];
		const token &tb = b[i];
		/* Assume tokensets haven't been merged yet, so we only have to worry about code and strings */
		if(ta.type != tb.type)
			return false;
		if((ta.type == ECODE) && (ta.value != tb.value))
			return false;
	}
	return true;
}

void merge_tokensets(const vector<tokenset> &tokensets,tokenset &result,vector<map<string,string> > &arguments)
{
	/* Merge a set of compatible tokensets, translating any differentiating strings into arguments */
	const tokenset &ref = tokensets[0];
	int numargs = 0;
	arguments.resize(tokensets.size());
	for(size_t i=0;i<ref.size();i++)
	{
		const token &tok = ref[i];
		token t = tok;
		if(t.type == ESTRING)
		{
			/* Check to see if this string differs in any other tokensets */
			for(size_t j=1;j<tokensets.size();j++)
			{
				const tokenset &other = tokensets[j];
				if(other[i].value != t.value)
				{
					/* Convert to parameter */
					char name[128];
					sprintf(name,"%sargs_%d",prefix,numargs++);
					t.type = ESTRINGREF;
					t.value = name;
					for(j=0;j<tokensets.size();j++)
					{
						const tokenset &other2 = tokensets[j];
						arguments[j][t.value] = other2[i].value;
					}
					break;
				}
			}
		}
		result.push_back(t);
	}
}
